import os
from subprocess import CompletedProcess


def check_output(subprocess: CompletedProcess, exit_code: int = 0, min_stdout_len=1) -> None:
    if not subprocess.returncode == exit_code and len(subprocess.stdout) > min_stdout_len:
        raise AssertionError('\n'.join([
            f'Command failed: {subprocess.args}',
            f'Exit code: {subprocess.returncode}',
            f'Stdout: {subprocess.stdout}',
            f'Stderr: {subprocess.stderr}']))


def get_paths() -> [str]:
    return os.environ['PATH'].split(os.pathsep)


def is_installed(program):
    """
    Test if a program is installed.

    :param program: path to executable or command
    :return: if program executable: program; else None
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:  # check if path to program is valid
        return is_exe(program)
    else:  # check if program is in PATH
        for path in get_paths():
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
        return False


def query_options(options: [str], question: str = 'Please select an option', error_msg='Please respond with a valid integer.') -> str:
    id_to_option = dict(enumerate(options, start=1))

    print('Options:')
    for id, path in id_to_option.items():
        print(f'  {id}: {path}')

    print(question)
    while True:
        choice = input()
        if choice.isdigit():
            choice = int(choice)
        if choice in id_to_option:
            result = id_to_option[choice]
            print(f'You chose: {result}')
            return result
        else:
            print(error_msg)



from functools import cached_property

from . import __expected_gendiscal_version__
import logging
import os
import numpy as np
import pandas as pd
from io import StringIO
from subprocess import run, PIPE
from .utils import is_installed, check_output

PACKAGE_ROOT = os.path.dirname(__file__)


class GenDisCal:
    def __init__(self, bin: str = None):
        self.bin = 'GenDisCal' if bin is None else bin
        assert is_installed(self.bin), f'GenDisCal binary is missing! {self.bin=}'
        if not self.version == __expected_gendiscal_version__:
            logging.warning(
                f'GenDisCal version does not match GenDisCal.py! expected={__expected_gendiscal_version__}; real=')

    @cached_property
    def version(self) -> str:
        """
        Get version of GenDisCal binary

        :return: version (string)
        """
        command = [self.bin, '--version']
        subprocess = run(command, stdout=PIPE, stderr=PIPE, encoding='ascii')
        check_output(subprocess)
        return subprocess.stdout.strip()

    def compare_two(self, assembly_1: str, assembly_2: str, preset: str = None, method: str = None) -> np.float64:
        # Throw error if files are not readable (see https://github.com/LM-UGent/GenDisCal/issues/2)
        for file in (assembly_1, assembly_2):
            if not os.access(file, os.R_OK):
                raise IOError(f'Cannot read assembly={file}')

        stdout = self._run(assembly_1, assembly_2, preset=preset, method=method)
        table = self._parse_table(stdout=stdout)

        return table.Distance.max()

    def run(
            self,
            *files: [str],
            file_list: bool = False,
            distance_matrix: bool = False,
            histogram: bool = False,
            preset: str = None,
            method: str = None
    ) -> pd.DataFrame:
        """
        Execute GenDisCal, return table/distance-matrix/histogram as pd.DataFrame

        :param files: list of fasta files
        :param file_list: (optional) file containing a list of paths to files to use
        :param preset: use a special preset of options (more info: run 'GenDisCal --preset -h')
        :param method:  distance calculation method to be used (more info: run 'GenDisCal --method -h')
        :param distance_matrix: if true, return distance matrix instead of table
        :param histogram:  if true, return histogram instead of table
        :return: table/distance-matrix/histogram as pd.DataFrame
        """
        stdout = self._run(*files, file_list=file_list, distance_matrix=distance_matrix, histogram=histogram,
                           preset=preset, method=method)

        if distance_matrix:
            return self._parse_dist(stdout)
        elif histogram:
            return self._parse_hist(stdout)
        else:
            return self._parse_table(stdout)

    def _run(
            self,
            *files: [str],
            file_list: bool = False,
            distance_matrix: bool = False,
            histogram: bool = False,
            preset: str = None,
            method: str = None
    ) -> str:
        """
        Execute GenDisCal, return table/distance-matrix/histogram raw string (stdout)

        :param files: list of fasta files
        :param file_list: (optional) file containing a list of paths to files to use
        :param preset: use a special preset of options (more info: run 'GenDisCal --preset -h')
        :param method:  distance calculation method to be used (more info: run 'GenDisCal --method -h')
        :param distance_matrix: if true, return distance matrix instead of table
        :param histogram:  if true, return histogram instead of table
        :return: table/distance-matrix/histogram as pd.DataFrame
        """
        assert files, f'No files were specified! {files=}'
        assert not (
                    distance_matrix and histogram), f'Error: options distance_matrix and histogram are mutually exclusive!'

        command = [self.bin]

        if type(preset) is str:
            command.extend(['--preset', preset])
        if type(method) is str:
            command.extend(['--method', method])

        if file_list:
            command.append('--filelist')

        command += list(files)

        if distance_matrix:
            command.append('--distancematrix')
        elif histogram:
            command.append('--histogram')

        subprocess = run(' '.join(command), stdout=PIPE, stderr=PIPE, encoding='ascii', shell=True)
        check_output(subprocess)

        logging.info('command=' + ' '.join(command))

        return subprocess.stdout.strip()

    @staticmethod
    def _parse_table(stdout: str) -> pd.DataFrame:
        res = pd.read_csv(
            StringIO(stdout), sep=',',
            dtype={'File1': str, 'File2': str, 'Expected_Relation': str, 'Distance': np.float64}
        )
        columns = res.columns.tolist()
        assert columns == ['File1', 'File2', 'Expected_Relation', 'Distance'], f'Columns not as expected: {columns}'
        return res

    @staticmethod
    def _parse_hist(stdout: str) -> pd.DataFrame:
        res = pd.read_csv(
            StringIO(stdout), sep=',',
            dtype={'bin': np.float64, '?': np.float64},
            index_col='bin'
        )
        columns = res.columns.tolist()
        assert columns == ['?'], f'Columns not as expected: {columns}'
        return res

    @staticmethod
    def _parse_dist(stdout: str) -> pd.DataFrame:
        res = pd.read_csv(
            StringIO(stdout), sep=',',
            index_col=0
        )
        assert res.values.dtype == np.float64, f'Table must contain only floats: {res.values.dtype}'
        assert list(res.columns) == list(res.index), 'Table must be symmetrical!'
        assert np.allclose(res, res.T), 'Table must be symmetrical!'
        return res


def main():
    from fire import Fire

    gdc = GenDisCal()
    Fire(gdc._run)


if __name__ == '__main__':
    main()



__version__ = '0.0.2'
__expected_gendiscal_version__ = 'GenDisCal v1.3.1'
__gendiscal_binary_url__ = 'https://github.com/LM-UGent/GenDisCal/releases/download/1.3.1/GenDisCal_v1_3_1_Linux.tar.gz'

from .GenDisCal import GenDisCal
from .GenDisCalTree import GenDisCalTree



from .GenDisCal import GenDisCal
from biotite.sequence.phylo import upgma


class GenDisCalTree:
    @staticmethod
    def from_csv(csv: str) -> str:
        """
        Create a phylogenetic tree based on GenDisCal a distance matrix.

        :param csv: Path to distance matrix
        :return: Newick string
        """
        with open(csv) as f:
            distance_matrix = GenDisCal._parse_dist(f.read())
        tree = upgma(distance_matrix.values)
        newick = tree.to_newick(labels=distance_matrix.index)
        return newick

    @staticmethod
    def from_files(*files: [str], file_list: bool = False, preset: str = None, method: str = None) -> str:
        """
        Create a phylogenetic tree directly based on assembly FASTA files.

        :param files: list of fasta files
        :param file_list: (optional) file containing a list of paths to files to use
        :param preset: use a special preset of options (more info: run 'GenDisCal --preset -h')
        :param method:  distance calculation method to be used (more info: run 'GenDisCal --method -h')
        :return: Newick string
        """
        distance_matrix = GenDisCal().run(*files, distance_matrix=True, file_list=file_list, preset=preset, method=method)
        tree = upgma(distance_matrix.values)
        newick = tree.to_newick(labels=distance_matrix.index)
        return newick


def main():
    from fire import Fire

    Fire({
        'from_csv': GenDisCalTree.from_csv,
        'from_files': GenDisCalTree.from_files
    })


if __name__ == '__main__':
    main()



import os
from urllib import request
import tarfile
from . import __gendiscal_binary_url__
from .utils import query_options, get_paths


def download_and_extract(target_path: str) -> str:
    temp_file, headers = request.urlretrieve(
        url=__gendiscal_binary_url__
    )

    with tarfile.open(temp_file) as tar:
        tar.extract(member='GenDisCal', path=target_path)

    os.remove(temp_file)

    gendiscal_bin = f'{target_path}/GenDisCal'

    assert os.access(gendiscal_bin, os.X_OK), f'GenDisCal binary is not executable! {gendiscal_bin=}'

    return gendiscal_bin


def select_path() -> str:
    path = query_options(options=get_paths(), question='Please select a path to install GenDisCal to!')
    return path


def install(path: str = None):
    if path is None:
        path = select_path()
    assert os.path.isdir(path), f'Error: Directory does not exist: {path=}'

    gendiscal_bin = download_and_extract(target_path=path)

    print(f'Installed GenDiscal here: {gendiscal_bin}')


def main():
    from fire import Fire

    Fire(install)


if __name__ == '__main__':
    main()



# based on https://realpython.com/pypi-publish-python-package
import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / 'README.md').read_text()

# This call to setup() does all the work
setup(
    name='gendiscalpy',
    version='0.0.2',
    description='Python bindings to GenDisCal and addition of phylogenetic tree function',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/MrTomRod/gendiscalpy',
    author='Thomas Roder',
    author_email='roder.thomas@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
    packages=['gendiscalpy'],
    install_requires=['fire', 'pandas', 'biotite'],
    entry_points={
        'console_scripts': [
            'install_gendiscal=gendiscalpy.install_gendiscal:main',
            'gendiscal_tree=gendiscalpy.GenDisCalTree:main',
        ]
    },
)
